/*
 * $Id: socket.c,v 1.3 1999/06/27 09:06:46 masaki Exp $
 */

#include <mrt.h>
#ifdef HAVE_NET_IF_DL_H
#include <net/if_dl.h>
#endif /* HAVE_NET_IF_DL_H */
#include <sys/uio.h>
#include <api6.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#ifdef HAVE_NETINET_ICMP6_H
#include <netinet/icmp6.h>
#endif /* HAVE_NETINET_ICMP6_H */


int
socket_open (int family, int type, int proto)
{
    int sockfd;

    if ((sockfd = socket (family, type, proto)) < 0) {
	trace (TR_ERROR, MRT->trace, "socket: %m\n");
    }
    return (sockfd);
}


int
socket_reuse (int sockfd, int yes)
{
    int ret;

    if ((ret = setsockopt (sockfd, SOL_SOCKET, SO_REUSEADDR, 
	    (char *) &yes, sizeof (yes))) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt SO_REUSEADDR (%d): %m\n", yes);
	return (ret);
    }
#ifdef SO_REUSEPORT
    if ((ret = setsockopt (sockfd, SOL_SOCKET, SO_REUSEPORT, 
	    (char *) &yes, sizeof (yes))) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt SO_REUSEPORT (%d): %m\n", yes);
    }
#endif /* SO_REUSEPORT */
    return (ret);
}


int
socket_broadcast (int sockfd, int yes)
{
    int ret;

#ifdef SO_BROADCAST
    if ((ret = setsockopt (sockfd, SOL_SOCKET, SO_BROADCAST, 
	    (char *) &yes, sizeof (yes))) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt SO_BROADCAST (%d): %m\n", yes);
	return (ret);
    }
#endif /* SO_BROADCAST */
    return (ret);
}


int
socket_rcvbuf (int sockfd, int size)
{
    int ret = -1;

#ifdef SO_RCVBUF
    if ((ret = setsockopt (sockfd, SOL_SOCKET, SO_RCVBUF, (char *) &size,
		    sizeof (size))) < 0) {
	trace (TR_ERROR, MRT->trace, "setsockopt SO_RCVBUF (%d): %m\n",
	       size);
    }
#endif /* SO_RCVBUF */
    return (ret);
}


int
ip_hdrincl (int sockfd, int yes)
{
    int ret = -1;

#ifdef IP_HDRINCL
    if ((ret = setsockopt (sockfd, IPPROTO_IP, IP_HDRINCL, (char *) &yes,
		    sizeof (yes))) < 0) {
	trace (TR_ERROR, MRT->trace, "setsockopt IP_HDRINCL (%d): %m\n",
	       yes);
    }
#endif /* IP_HDRINCL */
    return (ret);
}


int
ip_multicast_loop (int sockfd, int yes)
{
    u_char cyes = yes; /* char */
    int ret = -1;

#ifdef IP_MULTICAST_LOOP
    /* disable loopback */
    if ((ret = setsockopt (sockfd, IPPROTO_IP, IP_MULTICAST_LOOP,
	        &cyes, sizeof (cyes))) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt IP_MULTICAST_LOOP (%d): %s\n", 
	        cyes, strerror (errno));
    }
#endif /* IP_MULTICAST_LOOP */
    return (ret);
}


int
ip_multicast_hops (int sockfd, int hops)
{
    u_char ttl = hops;
    int ret = -1;

#ifdef IP_MULTICAST_HOPS
    if (setsockopt (sockfd, IPPROTO_IP, IP_MULTICAST_HOPS,
			(char *) &ttl, sizeof (ttl)) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt IP_MULTICAST_HOPS (%d): %s\n",
	       ttl, strerror (errno));
    }
#else /* IP_MULTICAST_HOPS */
#ifdef IP_MULTICAST_TTL
    if (setsockopt (sockfd, IPPROTO_IP, IP_MULTICAST_TTL,
			(char *) &ttl, sizeof (ttl)) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt IP_MULTICAST_TTL (%d): %s\n",
	       ttl, strerror (errno));
    }
#endif /* IP_MULTICAST_TTL */
#endif /* IP_MULTICAST_HOPS */
    return (ret);
}


int
ip_recvdstaddr (int sockfd, int yes)
{
    int ret = -1;

#if defined (IP_RECVDSTADDR) && !defined(sun)
    if ((ret = setsockopt (sockfd, IPPROTO_IP, IP_RECVDSTADDR, &yes,
		    sizeof (yes))) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt IP_RECVDSTADDR: %s\n", strerror (errno));
    }
#endif /* IP_RECVDSTADDR */
    return (ret);
}


int
ip_recvttl (int sockfd, int yes)
{
    int ret = -1;

#ifdef IP_UNICAST_HOPS
    if ((ret = setsockopt (sockfd, IPPROTO_IP, IP_UNICAST_HOPS, &yes,
		    sizeof (yes))) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt IP_UNICAST_HOPS (%d): %s\n", yes, strerror (errno));
    }
#else /* IP_UNICAST_HOPS */
#ifdef IP_RECVTTL
    if ((ret = setsockopt (sockfd, IPPROTO_IP, IP_RECVTTL, &yes,
		    sizeof (yes))) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt IP_RECVTTL (%d): %s\n", yes, strerror (errno));
    }
#endif /* IP_RECVTTL */
#endif /* IP_UNICAST_HOPS */
    return (ret);
}


int
ip_recvinterface (int sockfd, int yes)
{
    int ret = -1;

#ifdef IP_RECVINTERFACE
    if ((ret = setsockopt (sockfd, IPPROTO_IP, IP_RECVINTERFACE, &yes,
		    sizeof (yes))) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt IP_RECVINTERFACE (%d): %s\n", yes, strerror (errno));
    }
#else
#ifdef IP_RECVIF
    if ((ret = setsockopt (sockfd, IPPROTO_IP, IP_RECVIF, &yes,
		    sizeof (yes))) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt IP_RECVIF (%d): %s\n", yes, strerror (errno));
    }
#endif /* IP_RECVIF */
#endif /* IP_RECVINTERFACE */
    return (ret);
}


#ifdef HAVE_IPV6

/* These functions enable getting info on recvmsg() */
/* On Solaris, some may work, but it's unknown how to get info later */
int
ipv6_pktinfo (int sockfd, int yes)
{
    int ret = -1;

#ifdef IPV6_RECVPKTINFO
    /* XXX -- This is required by INRIA IPV6 to get pktinfo
       IPV6_PKTINFO does not work for setsockopt as RFC 2292 defines. */
    /*
     *  Receive destination address and interface from socket
     */
    if ((ret = setsockopt (sockfd, IPPROTO_IPV6, IPV6_RECVPKTINFO, &yes,
		    sizeof (yes))) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt IPV6_RECVPKTINFO: %s\n", strerror (errno));
    }
#else /* IPV6_RECVPKTINFO */
#ifdef IPV6_PKTINFO
    /*
     *  Receive the interface info from socket (RFC 2292)
     */
    if ((ret = setsockopt (sockfd, IPPROTO_IPV6, IPV6_PKTINFO, &yes,
		    sizeof (yes))) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt IPV6_PKTINFO: %s\n", strerror (errno));
    }
#else /* IPV6_PKTINFO */
#ifdef IP_RECVINTERFACE
    /* old INRIA IPv6 requires this */
    if (setsockopt (sockfd, IPPROTO_IPV6, IP_RECVINTERFACE, &yes,
		    sizeof (yes)) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt IP_RECVINTERFACE: %s\n", strerror (errno));
    }
#else
#ifdef IP_RECVIF
    if (setsockopt (sockfd, IPPROTO_IPV6, IP_RECVIF, &yes,
		    sizeof (yes)) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt IP_RECVIF: %s\n", strerror (errno));
    }
#endif /* IP_RECVIF */
#endif /* IP_RECVINTERFACE */
#endif /* IPV6_PKTINFO */
#endif /* IPV6_RECVPKTINFO */
    return (ret);
}


int
ipv6_recvdstaddr (int sockfd, int yes)
{
    int ret = -1;

#ifdef IPV6_RECVDSTADDR
    if ((ret = setsockopt (sockfd, IPPROTO_IPV6, IPV6_RECVDSTADDR, &yes,
		    sizeof (yes))) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt IPV6_RECVDSTADDR: %s\n", strerror (errno));
    }
#else /* IPV6_RECVDSTADDR */
#if defined(IP_RECVDSTADDR) && !defined(sun)
    if ((ret = setsockopt (sockfd, IPPROTO_IPV6, IP_RECVDSTADDR, &yes,
		    sizeof (yes))) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt IP_RECVDSTADDR: %s\n", strerror (errno));
    }
#endif /* IP_RECVDSTADDR */
#endif /* IPV6_RECVDSTADDR */
    return (ret);
}


int
ipv6_recvhops (int sockfd, int yes)
{
    int ret = -1;

#ifdef IPV6_RECVHOPS
    if ((ret = setsockopt (sockfd, IPPROTO_IPV6, IPV6_RECVHOPS, &yes,
		    sizeof (yes))) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt IPV6_RECVHOPS (%d): %s\n", yes, strerror (errno));
    }
#else /* IPV6_RECVHOPS */
#ifdef IPV6_HOPLIMIT
    if ((ret = setsockopt (sockfd, IPPROTO_IPV6, IPV6_HOPLIMIT, &yes,
		    sizeof (yes))) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt IPV6_HOPLIMIT (%d): %s\n", yes, strerror (errno));
    }
#endif /* IPV6_HOPLIMIT */
#endif /* IPV6_RECVHOPS */
    return (ret);
}


int
ipv6_unicast_hops (int sockfd, int hops)
{
    int ret = -1;
#ifndef NRL_IPV6
    int ttl = hops;
#else /* NRL_IPV6 */
/*
 * On NRL's IPV6, IPV6_MULTICAST_LOOP and IPV6_MULTICAST_HOPS
 * require a pointer to one byte (this was a log time ago...)
 */
    u_char ttl = hops;
#endif /* NRL_IPV6 */

#ifdef IPV6_UNICAST_HOPS
    if ((ret = setsockopt (sockfd, IPPROTO_IPV6, IPV6_UNICAST_HOPS, 
			  (char *) &ttl, sizeof (ttl))) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt IPV6_UNICAST_HOPS (%d): %s\n", 
	       ttl, strerror (errno));
    }
#endif /* IPV6_UNICAST_HOPS */
    return (ret);
}


int
ipv6_multicast_loop (int sockfd, int yes)
{
#ifndef NRL_IPV6
    int value = yes;
#else /* NRL_IPV6 */
    u_char value = yes;
#endif /* NRL_IPV6 */
    int ret = -1;
#ifdef IPV6_MULTICAST_LOOP
    if ((ret = setsockopt (sockfd, IPPROTO_IPV6, IPV6_MULTICAST_LOOP,
			(char *) &value, sizeof (value))) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt IPV6_MULTICAST_LOOP (%d): %s\n",
	       value, strerror (errno));
    }
#endif /* IPV6_MULTICAST_LOOP */
    return (ret);
}


int
ipv6_multicast_hops (int sockfd, int hops)
{
#ifndef NRL_IPV6
    int ttl = hops;
#else /* NRL_IPV6 */
    u_char ttl = hops;
#endif /* NRL_IPV6 */
    int ret = -1;

#ifdef IPV6_MULTICAST_HOPS
    if (setsockopt (sockfd, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
			(char *) &ttl, sizeof (ttl)) < 0) {
	trace (TR_ERROR, MRT->trace,
	       "setsockopt IPV6_MULTICAST_HOPS (%d): %s\n",
	       ttl, strerror (errno));
    }
#endif /* IPV6_MULTICAST_HOPS */
    return (ret);
}


int
ipv6_hdrincl (int sockfd, int yes)
{
    int ret = -1;

#ifdef IP_HDRINCL
    if ((ret = setsockopt (sockfd, IPPROTO_IPV6, IP_HDRINCL, (char *) &yes,
		    sizeof (yes))) < 0) {
	trace (TR_ERROR, MRT->trace, "setsockopt IP_HDRINCL (%d): %s\n",
	       yes, strerror (errno));
    }
#endif /* IP_HDRINCL */
    return (ret);
}


#ifdef ICMP6_FILTER
int
icmp6_filter (int sockfd, struct icmp6_filter *filter)
{
	int ret = -1;

    if ((ret = setsockopt (sockfd, IPPROTO_ICMPV6, ICMP6_FILTER, (char *)filter,
		    sizeof (*filter))) < 0) {
	trace (TR_ERROR, MRT->trace, "setsockopt ICMP6_FILTER: %s\n",
	       strerror (errno));
    }
    return (ret);
}
#endif /* ICMP6_FILTER */

#endif /* HAVE_IPV6 */


int
join_leave_group (int sockfd, interface_t *interface, prefix_t *prefix, 
		  int join)
{
    struct ip_mreq mreq;
    char *mreqptr;
    int mreqlen;
#ifdef HAVE_IPV6
    struct ipv6_mreq mreq6;
#endif /* HAVE_IPV6 */
    int cmd;
    char *s_cmd;
    int proto;

    if (!BIT_TEST (interface->flags, IFF_UP))
	return (-1);
    if (!BIT_TEST (interface->flags, IFF_MULTICAST))
	return (-1);

    if (prefix->family == AF_INET) {
        if (interface->primary == NULL)
	    return (-1);

        memset (&mreq, 0, sizeof (mreq));
        mreq.imr_multiaddr.s_addr = prefix_tolong (prefix);
        mreq.imr_interface.s_addr = prefix_tolong (interface->primary->prefix);
        mreqptr = (char *) &mreq;
	mreqlen = sizeof (mreq);

        if (join) {
            cmd = IP_ADD_MEMBERSHIP;
            s_cmd = "IP_ADD_MEMBERSHIP";
 	}
	else {
            cmd = IP_DROP_MEMBERSHIP;
            s_cmd = "IP_DROP_MEMBERSHIP";
	}
	proto = IPPROTO_IP;
    }
#ifdef HAVE_IPV6
    else if (prefix->family == AF_INET6) {
        if (interface->primary6 == NULL)
	    return (-1);
        memset (&mreq6, 0, sizeof (mreq6));
        memcpy (&mreq6.ipv6mr_multiaddr, prefix_toaddr6 (prefix), 16);

#ifdef __KAME__
        /* On KAME IPv6, it is required to modify the destination address
           as to include the outgoing interface's index */
        /* This is unnessesary when responding to a unicast address which
           includes an index */

        mreq6.ipv6mr_multiaddr.s6_addr8[2] = (interface->index >> 8) & 0xff;
        mreq6.ipv6mr_multiaddr.s6_addr8[3] = interface->index & 0xff;
#endif /* __KAME__ */

#if defined (NRL_IPV6) || defined (SOLARIS_IPV6)
        memcpy (&mreq6.ipv6mr_interface,
	        prefix_toaddr6 (interface->primary6->prefix), 16);
#else
        mreq6.ipv6mr_interface = interface->index;
#endif /* NRL_IPV6 && SOLARIS_IPV6 */
        mreqptr = (char *) &mreq6;
	mreqlen = sizeof (mreq6);

        if (join) {
#ifdef IPV6_JOIN_MEMBERSHIP
            cmd = IPV6_JOIN_MEMBERSHIP;
            s_cmd = "IPV6_JOIN_MEMBERSHIP";
#else
            cmd = IPV6_ADD_MEMBERSHIP;
            s_cmd = "IPV6_ADD_MEMBERSHIP";
#endif /* IPV6_JOIN_MEMBERSHIP */
        }
	else {
#ifdef IPV6_LEAVE_MEMBERSHIP
            cmd = IPV6_LEAVE_MEMBERSHIP;
            s_cmd = "IPV6_LEAVE_MEMBERSHIP";
#else
            cmd = IPV6_DROP_MEMBERSHIP;
            s_cmd = "IPV6_DROP_MEMBERSHIP";
#endif /* IPV6_LEAVE_MEMBERSHIP */
	}
        proto = IPPROTO_IPV6;
    }
#endif /* HAVE_IPV6 */
    else
	return (-1);

    if (setsockopt (sockfd, proto, cmd, mreqptr, mreqlen) >= 0) {
	trace (TR_TRACE, MRT->trace, "%s %s on %s\n", 
	       s_cmd, prefix_toa (prefix), interface->name);
    }
    else {
	trace (TR_ERROR, MRT->trace, "setsockopt %s %s for %s: %s\n", 
	       s_cmd, prefix_toa (prefix), interface->name, strerror (errno));
	    return (-1);
    }
    return (0);
}


int
ip_multicast_if (int sockfd, interface_t *interface) 
{
    int ret = -1;

    assert (interface);
    if (interface->primary == NULL)
	return (ret);
    assert (interface->primary->prefix);
    if ((ret = setsockopt (sockfd, IPPROTO_IP, IP_MULTICAST_IF,
                    prefix_tochar (interface->primary->prefix),
                    sizeof (struct in_addr))) < 0) {
            trace (TR_ERROR, MRT->trace,
                   "setsockopt IP_MULTICAST_IF for %s: %s\n",
                   interface->name, strerror (errno));
    }
    return (ret);
}


int
send_packet (int sockfd, char *msg, int len, int flags, 
	     prefix_t *prefix, int port, interface_t *interface, int flowinfo)
{
    int ret = -1;
    int mc;

    assert (sockfd > 0);
    assert (msg);
    assert (len > 0);
    assert (prefix);

    mc = prefix_is_multicast (prefix);

    if (interface != NULL && !BIT_TEST (interface->flags, IFF_TUNNEL)) {

	if (!BIT_TEST (interface->flags, IFF_UP))
            return (ret);
        if (prefix->family == AF_INET && interface->primary == NULL)
            return (ret);
#ifdef HAVE_IPV6
        if (prefix->family == AF_INET6 && interface->primary6 == NULL)
            return (ret);
#endif /* HAVE_IPV6 */

	if (mc) {
            if (!BIT_TEST (interface->flags, IFF_MULTICAST)) {
		prefix_t *old_prefix = prefix;

		/* fall back to unicast */
		/* should I have a flag to allow this on interface->flags ? */
                if (BIT_TEST (interface->flags, IFF_POINTOPOINT)) {
#ifdef HAVE_IPV6
		    if (prefix->family == AF_INET6 && interface->link_local->broadcast) {
		        prefix = interface->link_local->broadcast;
                        trace (TR_TRACE, MRT->trace, 
		    "send using %a (p2p dest) in place of %a (multicast)\n", 
		               prefix, old_prefix);
		    }
		    else
#endif /* HAVE_IPV6 */
		    if (interface->primary->broadcast) {
		        prefix = interface->primary->broadcast;
                        trace (TR_TRACE, MRT->trace, 
		    "send using %a (p2p dest) in place of %a (multicast)\n", 
		               prefix, old_prefix);
		    }
		    else
                        return (ret);
		}
                else if (BIT_TEST (interface->flags, IFF_BROADCAST) 
			&& interface->primary) {
		    /* no broadcast in IPv6 */
		    prefix = interface->primary->broadcast;
                    trace (TR_TRACE, MRT->trace, 
		"send using %a (broadcast) in place of %a (multicast)\n", 
		           prefix, old_prefix);
		}
		else
                    return (ret);
	    }
	}
    }

    if (prefix->family == AF_INET) {
	struct sockaddr_in sin;

	if (interface && mc) {
	    if (ip_multicast_if (sockfd, interface) < 0)
	        return (ret);
	}

	memset (&sin, 0, sizeof (sin));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = prefix_tolong (prefix);
	sin.sin_port = htons (port);
#ifdef HAVE_SA_LEN
	sin.sin_len = sizeof (sin);
#endif /* HAVE_SA_LEN */

	if ((ret = sendto (sockfd, msg, len, flags,
                     (struct sockaddr *) &sin, sizeof (sin))) < 0) {
            trace (TR_ERROR, MRT->trace, "sendto %s on %s: %s\n",
                   prefix_toa (prefix), (interface)?interface->name:"?",
		   strerror (errno));
        }
        else {
            trace (TR_TRACE, MRT->trace, "send %d bytes to %a on %s\n", 
		   len, prefix, (interface)?interface->name:"?");
        }
    }
#ifdef HAVE_IPV6
    else if (prefix->family == AF_INET6) {
	struct sockaddr_in6 sin6;
#ifdef IPV6_PKTINFO
        struct msghdr mhdr;
        struct cmsghdr *cmsg;
        struct in6_pktinfo *pkt_info; 
        struct iovec iov;
        char chdr[sizeof (struct cmsghdr) + sizeof (struct in6_pktinfo)];
#endif /* IPV6_PKTINFO */

	memset (&sin6, 0, sizeof (sin6));
	sin6.sin6_family = AF_INET6;
	memcpy (&sin6.sin6_addr, prefix_toaddr6 (prefix), 16);
	sin6.sin6_flowinfo = htonl (flowinfo);
	sin6.sin6_port = htons (port);
#ifdef HAVE_SA_LEN
	sin6.sin6_len = sizeof (sin6);
#endif /* HAVE_SA_LEN */

#ifndef IPV6_PKTINFO
	if (interface && mc) {
	    if (ipv6_multicast_if (sockfd, interface) < 0)
	        return (ret);
	}
	if ((ret = sendto (sockfd, msg, len, flags,
                     (struct sockaddr *) &sin6, sizeof (sin6))) < 0) {
            trace (TR_ERROR, MRT->trace, "sendto %a on %s: %m\n",
                   prefix, (interface)?interface->name:"?");
        }
        else {
            trace (TR_TRACE, MRT->trace, "send %d bytes to %a on %s\n", 
		   len, prefix, (interface)?interface->name:"?");
        }
#else
        /* RFC 2292 */
#ifdef __KAME__
        /* I'm not sure the recent KAME requires this.
           I'm not sure in which ways this is required */

	if (interface) {
            /* On KAME IPv6, it is required to modify the destination address
               as to include the outgoing interface's index */

            sin6.sin6_addr.s6_addr8[2] = (interface->index >> 8) & 0xff;
            sin6.sin6_addr.s6_addr8[3] = interface->index & 0xff;
        }
#endif /* __KAME__ */
	memset (&mhdr, 0, sizeof (mhdr));
        mhdr.msg_name = (char *) &sin6;
        mhdr.msg_namelen = sizeof (sin6);
	memset (&iov, 0, sizeof (iov));
        iov.iov_base = msg;
        iov.iov_len = len;
        mhdr.msg_iov = &iov;
        mhdr.msg_iovlen = 1;
	memset (&chdr, 0, sizeof (chdr));
        cmsg = (struct cmsghdr *) chdr;
        cmsg->cmsg_len = sizeof (chdr);
        cmsg->cmsg_level = IPPROTO_IPV6;
        cmsg->cmsg_type = IPV6_PKTINFO;
        mhdr.msg_control = (void *) cmsg; 
        mhdr.msg_controllen = sizeof (chdr);
        pkt_info = (struct in6_pktinfo *) CMSG_DATA (cmsg);
	if (interface) {
            pkt_info->ipi6_ifindex = interface->index;
	    if (prefix_is_linklocal (prefix) && interface->link_local) {
                memcpy (&pkt_info->ipi6_addr,
                        prefix_toaddr6 (interface->link_local->prefix), 16);
		trace (TR_PACKET, MRT->trace,
                   "source address should be %a\n",
                   interface->link_local->prefix);
	    }
	}
	if ((ret = sendmsg (sockfd, &mhdr, flags)) < 0) {
            trace (TR_ERROR, MRT->trace, "sendmsg %a on %s: %m\n",
                   prefix, (interface)?interface->name:"?");
        }
        else {
            trace (TR_TRACE, MRT->trace, "sendmsg %d bytes to %a on %s\n", 
		   len, prefix, (interface)?interface->name:"?");
        }
#endif /* IPV6_PKTINFO */
    }
#endif /* HAVE_IPV6 */
    return (ret);
}


union sockunion {
    struct sockaddr sa;
    struct sockaddr_in sin;
#ifdef HAVE_IPV6
    struct sockaddr_in6 sin6;
#endif                          /* HAVE_IPV6 */
};


int
recvmsgfrom (int sockfd, char *buffer, int buflen, int flags,
	     prefix_t ** prefix_p, int *port_p, interface_t ** interface_p)
{
    int ret = -1;
    prefix_t *prefix = NULL;
    interface_t *interface = NULL;
    int index = 0;
    int port;
    union sockunion from;
#ifdef sun
    int fromlen = sizeof (from);
#else
    union sockunion dstaddr;
    int limit = 0;
#ifdef INET6_ADDRSTRLEN
    char tmp6[INET6_ADDRSTRLEN];
#else
    char tmp6[128];
#endif /* HAVE_IPV6 */

    struct msghdr mhdr;
    struct cmsghdr *cmsg;
#ifdef IPV6_PKTINFO
    struct in6_pktinfo *pkt_info;
#endif
    struct iovec iov;
    char chdr[1024];
#endif /* sun */
    int mayigmp;
    int dontcare;

    mayigmp = BIT_TEST (flags, MSG_MAYIGMP);
    BIT_RESET (flags, MSG_MAYIGMP);
    dontcare = BIT_TEST (flags, MSG_DONTCARE);
    BIT_RESET (flags, MSG_DONTCARE);

#ifndef sun
    mhdr.msg_name = (void *) &from;
    mhdr.msg_namelen = sizeof (from);
    memset (mhdr.msg_name, 0, mhdr.msg_namelen);
    mhdr.msg_iov = &iov;
    mhdr.msg_iovlen = 1;
    mhdr.msg_control = (void *) chdr;
    mhdr.msg_controllen = sizeof (chdr);
    memset (mhdr.msg_control, 0, mhdr.msg_controllen);
    iov.iov_base = buffer;
    iov.iov_len = buflen;

    ret = recvmsg (sockfd, &mhdr, flags);
    if (ret <= 0) {
	/* zero is error? XXX */
	trace (TR_ERROR, MRT->trace, "recvmsg: %m\n");
	return (ret);
    }
    else {
	memset (&dstaddr, 0, sizeof (dstaddr));
/* fprintf(stderr, "mhdr.msg_control = 0x%x mhdr.msg_controllen = %d\n", 
   mhdr.msg_control, mhdr.msg_controllen); */
#ifdef __GLIBC__
/* There is a bug in GNU LIBC 2.0, so I define another one. */
#undef CMSG_NXTHDR
#define CMSG_NXTHDR(mhdr, cmsg) \
    (((unsigned char *)(cmsg) + (cmsg)->cmsg_len + sizeof(struct cmsghdr) > \
	(unsigned char *)(mhdr)->msg_control + (mhdr)->msg_controllen) ? \
	(struct cmsghdr *)NULL : \
        (struct cmsghdr *)((unsigned char *)(cmsg) \
         + (((cmsg)->cmsg_len + sizeof(long int)-1) & ~(sizeof(long int)-1))))
#endif /* __GLIBC__ */
	if (mhdr.msg_controllen > 0)
	    for (cmsg = CMSG_FIRSTHDR (&mhdr);
		 cmsg;
		 cmsg = CMSG_NXTHDR (&mhdr, cmsg)) {
/* fprintf(stderr, 
   "cmg=0x%x cmsg->cmsg_level = %d cmsg->cmsg_type = %d cmsg->cmsg_len = %d\n", 
   cmsg, cmsg->cmsg_level, cmsg->cmsg_type, cmsg->cmsg_len); */
		switch (cmsg->cmsg_type) {
#ifdef IPV6_PKTINFO
		case IPV6_PKTINFO:
		    pkt_info = (struct in6_pktinfo *) CMSG_DATA (cmsg);
		    index = pkt_info->ipi6_ifindex;
		    memcpy (&dstaddr.sin6.sin6_addr, &pkt_info->ipi6_addr, 16);
		    dstaddr.sa.sa_family = AF_INET6;
		    trace (TR_PACKET, MRT->trace,
			   "recv IPV6_PKTINFO dest = %s index = %d\n",
			   inet_ntop (AF_INET6, &dstaddr.sin6.sin6_addr,
				      tmp6, sizeof tmp6), index);
		    break;
#endif /* IPV6_PKTINFO */
#ifdef IPV6_RECVINTERFACE
		case IPV6_RECVINTERFACE:
		    index = *(u_short *) CMSG_DATA (cmsg);
		    trace (TR_PACKET, MRT->trace,
			   "recv IPV6_RECVINTERFACE index = %d\n", index);
		    break;
#endif /* IPV6_RECVINTERFACE */
#ifdef IP_RECVIF
		case IP_RECVIF:
{
		    struct sockaddr_dl *sdl;

		    sdl = (struct sockaddr_dl *) CMSG_DATA (cmsg);
		    index = sdl->sdl_index;
		    trace (TR_PACKET, MRT->trace,
			   "recv IP_RECVIF index = %d\n", index);
		    break;
}
#endif /* IP_RECVIF */
#ifdef IP_RECVINTERFACE
		case IP_RECVINTERFACE:
		    index = *(u_short *) CMSG_DATA (cmsg);
		    trace (TR_PACKET, MRT->trace,
			   "recv IP_RECVINTERFACE index = %d\n", index);
		    break;
#endif /* IP_RECVINTERFACE */
		    /*
		     * XXX destination address of packet should be checked 
		     * RIPNG request may not be addressed with multicast addess
		     * it may be possible later
		     */
#ifdef IPV6_RECVDSTADDR
		case IPV6_RECVDSTADDR:
		    dstaddr.sa.sa_family = AF_INET6;
		    memcpy (&dstaddr.sin6.sin6_addr, CMSG_DATA (cmsg), 16);
		    trace (TR_PACKET, MRT->trace,
			   "recv IPV6_RECVDSTADDR dest = %s\n",
			 inet_ntop (AF_INET6, &dstaddr.sin6.sin6_addr, 
				    tmp6, sizeof tmp6));
		    break;
#endif /* IPV6_RECVDSTADDR */
#ifdef IP_RECVDSTADDR
		case IP_RECVDSTADDR:
		    dstaddr.sa.sa_family = from.sa.sa_family;
		    /* I could use cmsg_level, but it has a bug in the kernel
		       of INRIA_IPV6 */
#ifdef HAVE_IPV6
		    if (from.sa.sa_family == AF_INET6) {
		        memcpy (&dstaddr.sin6.sin6_addr, CMSG_DATA (cmsg), 16);
		        trace (TR_PACKET, MRT->trace,
			       "recv IP_RECVDSTADDR dest = %s\n",
			        inet_ntop (AF_INET6,
				           &dstaddr.sin6.sin6_addr, 
				           tmp6, sizeof tmp6));
		    }
		    else
#endif /* HAVE_IPV6 */
{
		    memcpy (&dstaddr.sin.sin_addr, CMSG_DATA (cmsg), 4);
		    trace (TR_PACKET, MRT->trace,
			   "recv IP_RECVDSTADDR dest = %s\n",
			    inet_ntop (AF_INET,
				       &dstaddr.sin.sin_addr, 
				       tmp6, sizeof tmp6));
}
		    break;
#endif /* IP_RECVDSTADDR */
		    /*
		     * XXX it may be possible later checking hop limit
		     */
#ifdef IPV6_HOPLIMIT
		case IPV6_HOPLIMIT:
		    limit = *(u_char *) CMSG_DATA (cmsg);
		    trace (TR_PACKET, MRT->trace,
			   "recv IPV6_HOPLIMIT = %d\n", limit);
		    break;
#endif /* IPV6_HOPLIMIT */
#ifdef IP_UNICAST_HOPS
		case IP_UNICAST_HOPS:
		    limit = *(int *) CMSG_DATA (cmsg);
		    trace (TR_PACKET, MRT->trace,
			   "recv IP_UNICAST_HOPS = %d\n", limit);
		    break;
#endif /* IP_UNICAST_HOPS */
#if defined(IPV6_UNICAST_HOPS) && IP_UNICAST_HOPS != IPV6_UNICAST_HOPS
		case IPV6_UNICAST_HOPS:
		    limit = *(int *) CMSG_DATA (cmsg);
		    trace (TR_PACKET, MRT->trace,
			   "recv IPV6_UNICAST_HOPS = %d\n", limit);
		    break;
#endif /* IPV6_UNICAST_HOPS */
#ifdef IP_RECVTTL
		case IP_RECVTTL:
		    limit = *(int *) CMSG_DATA (cmsg);
		    trace (TR_PACKET, MRT->trace,
			   "recv IP_RECVTTL = %d\n", limit);
		    break;
#endif /* IP_RECVTTL */
		default:
		    trace (TR_PACKET, MRT->trace,
			"recv unknown cmsg level = %d type = %d len = %d\n",
			 cmsg->cmsg_level, cmsg->cmsg_type, cmsg->cmsg_len);
		    break;
		}
	    }
    }
#else /* sun */

/* Solaris has IP_RECVDSTADDR, but differs. No way to know the interface */
    if ((ret = recvfrom (sockfd, (char *) buffer, buflen, flags,
			(struct sockaddr *) &from, &fromlen)) <= 0) {
	trace (TR_ERROR, MRT->trace, "recvfrom: %m\n");
	return (ret);
    }
#endif /* sun */

    if (prefix_p == NULL && interface_p == NULL)
	return (ret);

#ifdef HAVE_IPV6
    if (from.sa.sa_family == AF_INET6) {
        prefix = New_Prefix (AF_INET6, &from.sin6.sin6_addr, 128);
	port = ntohs (from.sin6.sin6_port);
    }
    else
#endif /* HAVE_IPV6 */
{
    prefix = New_Prefix (AF_INET, &from.sin.sin_addr, 32);
    port = ntohs (from.sin.sin_port);
}

    if (interface_p == NULL) {
        assert (prefix_p);
        *prefix_p = prefix;
        assert (port_p);
	*port_p = port;
        return (ret);
    }

    if (mayigmp && ((struct ip *) buffer)->ip_p == 0) {
	/* IGMP kernel message */
	dontcare++;
    }

    if (index > 0) {
	interface = find_interface_byindex (index);
	/* even in this case, we need to check the source
	   since it may come from through a tunnel */
	if (interface && !is_prefix_on (prefix, interface)) {
	    interface = find_tunnel_interface (prefix);
	    if (interface != NULL) {
	        trace (TR_PACKET, MRT->trace,
		   "recv tunnel source %a on %s\n",
		   prefix, interface->name);
	    }
	}
    }

    if (interface == NULL && dontcare == 0) {
	gateway_t *gateway;
	/*
	 * if the system doesn't supply the incoming interface index,
	 * try to find among configured interfaces and gateways registered,
	 */
	if ((interface = find_interface (prefix))) {
	    trace (TR_PACKET, MRT->trace,
		   "guesses %a on %s (find_interface)\n",
		   prefix, interface->name);
	}
	else if ((gateway = find_gateway (prefix)) &&
		 gateway->interface) {

	    interface = gateway->interface;
	    trace (TR_PACKET, MRT->trace,
		   "guesses %a on %s (find_gateway)\n",
		   prefix, interface->name);
	}
    }

    if (interface == NULL && dontcare == 0) {
	trace (TR_PACKET, MRT->trace,
	       "discard packet from %a (interface unknown)\n",
	       prefix);
	ret = -1;
    }
    assert (interface_p);
    *interface_p = interface;
    if (prefix_p) {
        *prefix_p = prefix;
	assert (port_p);
        *port_p = port;
    }
    else
        Deref_Prefix (prefix);
    return (ret);
}


#ifdef HAVE_IPV6
int
ipv6_multicast_if (int sockfd, interface_t *interface) 
{
    int ret = -1;
#if defined (__KAME__) || defined (INRIA_IPV6) || defined (__linux__)
    int index;
#endif

    assert (interface);
    if (interface->primary6 == NULL)
	return (ret);
    assert (interface->primary6->prefix);
#if defined (__KAME__) || defined (INRIA_IPV6) || defined (__linux__)
    index = interface->index;
    /* IPV6_PKTINFO will be used instead */
    if ((ret = setsockopt (sockfd, IPPROTO_IPV6, IPV6_MULTICAST_IF,
                    &index, sizeof (int))) < 0) {
#else
    if ((ret = setsockopt (sockfd, IPPROTO_IPV6, IPV6_MULTICAST_IF,
                    prefix_tochar (interface->primary6->prefix),
                    sizeof (struct in6_addr))) < 0) {
#endif
            trace (TR_ERROR, MRT->trace,
                   "setsockopt IPV6_MULTICAST_IF for %s: %m\n",
                   interface->name);
    }
    return (ret);
}
#endif /* HAVE_IPV6 */

